@@ -67,15 +67,15 @@
 #include "httpd.h"
 #include "http_config.h"
 #include "http_log.h"
+#include "http_main.h"
 
 /* Note that APACHE_SSL_EXPORT_CERTS is defined in buff.h */
 
-/*
-Useful bits for dbx (useful for Ben, at least):
+/* Temp */
+#define RENEG		FALSE
+#define DEBUG_RENEG	FALSE
 
-r -d /u/ben/apache/apache_1.2.5-ssl/SSLconf -X
-use /u/ben/work/scuzzy-ssleay8/ssl /u/ben/work/scuzzy-ssleay8/crypto/bio
-*/
+#define USE_OLD_ENVIRONMENT	FALSE	/* Set to TRUE to use the old SSL_{CLIENT,SERVER}_I* instead of SSL_{CLIENT,SERVER}_I_* */
 
 #define SSL_IS_OPTIONAL		FALSE	/* Set to TRUE if you want to allow SSL to be disabled on a per-request basis (useful for subrequests) */
 
@@ -160,8 +160,10 @@ typedef struct
     char *szCACertificatePath;
     char *szCACertificateFile;
     char *szReqCiphers;
+#if !RENEG
     int nVerifyDepth;
     VerifyType nVerifyClient;
+#endif
 
     X509 *px509Certificate;
     RSA *prsaKey;
@@ -184,6 +186,10 @@ typedef struct
     BOOL bSSLRequired;
 #if APACHE_SSL_EXPORT_CERTS
     BOOL bExportCerts;
+#endif
+#if RENEG
+    int nVerifyDepth;
+    VerifyType nVerifyClient;
 #endif
     } SSLDirConfigRec;
 
@@ -191,6 +197,9 @@ MODULE_VAR_EXPORT module apache_ssl_module;
 
 static conn_rec *s_pCurrentConnection;
 static pool *s_pPermanentPool;
+#if RENEG
+static SSLDirConfigRec *s_pConfig;
+#endif
 
 static const char six2pr[64+1]=
   "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
@@ -230,6 +239,179 @@ static void uuencode(char *szTo,const char *szFrom,BOOL bPad)
 	      strlen(szFrom),bPad);
     }
 
+static void SSLLogError(server_rec *s,int nLevel,const char *szErr)
+    {
+    unsigned long l;
+    char buf[MAX_STRING_LEN];
+
+    if(szErr)
+	ap_log_error(APLOG_MARK,nLevel|APLOG_NOERRNO,s,"%s",szErr);
+    while((l=ERR_get_error()))
+	{
+	ERR_error_string(l,buf);
+	ap_log_error(APLOG_MARK,nLevel|APLOG_NOERRNO,s,"%s",buf);
+	}
+    }
+
+/* for debugging */
+static void SSLLogErrorToFile(FILE *f)
+    {
+    unsigned long l;
+    char buf[MAX_STRING_LEN];
+
+    /* Print out error messages */
+
+    while((l=ERR_get_error()))
+	{
+	ERR_error_string(l,buf);
+	fputs(buf,f);
+	fputc('\n',f);
+	}
+    }
+
+#if APACHE_SSL_EXPORT_CERTS
+/* szEnv must have been pool allocated, or be static */
+static void AddCertToEnv(pool *p,table *e,char *szEnv,X509 *xs)
+    {
+    int n;
+    unsigned char *s,*d;
+
+    n=i2d_X509(xs,NULL);
+    d=s=ap_palloc(p,n);
+    i2d_X509(xs,&d);
+    d=ap_palloc(p,(n*4)/3+2+2);
+    uuencoden(d,s,n,TRUE);
+    ap_table_setn(e,szEnv,(char *)d);
+    }
+#endif
+
+#if SSLEAY_VERSION_NUMBER >= 0x0800
+static int ApacheSSLVerifyCallback(int ok,X509_STORE_CTX *ctx)
+    {
+    X509 *xs=X509_STORE_CTX_get_current_cert(ctx);
+    int depth=X509_STORE_CTX_get_error_depth(ctx);
+    int error=X509_STORE_CTX_get_error(ctx);
+#else
+static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
+				   int error)
+    {
+#endif
+#if RENEG
+    SSLDirConfigRec *pConfig=s_pConfig;
+#else
+    SSLConfigRec *pConfig=ap_get_module_config(s_pCurrentConnection->server->module_config,&apache_ssl_module);
+#endif
+    server_rec *pServer=s_pCurrentConnection->server;
+
+#if DEBUG_VERIFY
+      {
+      char *s;
+
+#if SSLEAY_VERSION_NUMBER < 0x0800
+      s=X509_NAME_oneline(X509_get_subject_name(xs));
+#else
+      s=X509_NAME_oneline(X509_get_subject_name(xs),NULL,0);
+#endif
+
+      if(s == NULL)
+	  {
+	  SSLLogError(pServer,APLOG_ERR,"can't get subject name");
+	  return(0);
+	  }
+
+      ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"depth=%d %s",
+		   depth,s);
+      free(s);
+      }
+#endif
+    
+#if APACHE_SSL_KEEP_CERTS
+    if(depth)
+	{
+	char *t;
+	pool *p=s_pCurrentConnection->client->pool;
+
+	if(!s_pCurrentConnection->client->tableCertChain)
+	    s_pCurrentConnection->client->tableCertChain=ap_make_table(p,5);
+
+	t=ap_psprintf(p,ENV_CERT_CHAIN "%d",depth);
+	AddCertToEnv(p,s_pCurrentConnection->client->tableCertChain,t,xs);
+	}
+#endif
+
+    if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
+       || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE
+       || error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
+       || error == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN)
+	{
+	char *s;
+
+	if(pConfig->nVerifyClient == VERIFY_OPTIONAL_NO_CA)
+	    {
+	    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+			 "no issuer, returning OK");
+	    return TRUE;
+	    }
+#if SSLEAY_VERSION_NUMBER < 0x0800
+	s=X509_NAME_oneline(X509_get_issuer_name(xs));
+#else
+	s=X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0);
+#endif
+	if(s == NULL)
+	    {
+	    SSLLogError(pServer,APLOG_ERR,"verify error");
+	    return FALSE;
+	    }
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"issuer=%s",
+		     s);
+	free(s);
+	}
+    if(!ok)
+	{
+	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,pServer,
+		      "verify error:num=%d:%s",error,
+		      X509_verify_cert_error_string(error));
+	SSLLogError(s_pCurrentConnection->server,APLOG_ERR,NULL);
+	s_pCurrentConnection->client->szClientX509=NULL;
+	}
+    if(depth >= pConfig->nVerifyDepth)
+	{
+	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,
+		     pServer,"Verify depth exceeded");
+	ok=0;
+	}
+    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+		 "verify return:%d",ok);
+    return ok;
+    }
+
+#if RENEG
+static int VerifyFlags(SSLDirConfigRec *pConfig)
+#else
+static int VerifyFlags(SSLConfigRec *pConfig)
+#endif
+    {
+    int nVerify=0;
+
+    switch(pConfig->nVerifyClient)
+	{
+    case VERIFY_REQUIRE:
+	nVerify|=SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
+	break;
+
+    case VERIFY_OPTIONAL:
+    case VERIFY_OPTIONAL_NO_CA:
+	nVerify|=SSL_VERIFY_PEER;
+	break;
+
+    default: break; /* avoid warnings from gcc */
+	}
+
+fprintf(stderr,"flags=%x\n",nVerify);
+
+    return nVerify;
+    }
+
 static int SSLCheckCipher(request_rec *r)
     {
     char *cipher;
@@ -237,6 +419,9 @@ static int SSLCheckCipher(request_rec *r)
       ap_get_module_config(r->per_dir_config,&apache_ssl_module);
     SSLConfigRec *pConfig=(SSLConfigRec *)
       ap_get_module_config(r->server->module_config,&apache_ssl_module);
+#if RENEG
+    int n;
+#endif
 
     if(rec->bSSLRequired && !r->connection->client->ssl)
 	{
@@ -248,6 +433,43 @@ static int SSLCheckCipher(request_rec *r)
     if(pConfig->bDisabled || OPTIONAL_SSL)
 	return DECLINED;
 
+#if RENEG
+# if DEBUG_RENEG
+    fprintf(stderr,"Renegotiated for %s (%x)!\n",r->filename,VerifyFlags(rec));
+# endif
+    if(1)
+	{
+	s_pConfig=rec;
+	SSL_set_verify(r->connection->client->ssl,VerifyFlags(rec),
+		       ApacheSSLVerifyCallback);
+	SSL_renegotiate(r->connection->client->ssl);
+	SSL_do_handshake(r->connection->client->ssl);
+	SSL_write(r->connection->client->ssl,NULL,0);
+	SSL_read(r->connection->client->ssl,NULL,0);
+	}
+
+    if(rec->nVerifyClient != VERIFY_NONE && rec->nVerifyClient != VERIFY_UNSET)
+	{
+	char *s;
+	X509 *xs;
+	xs=SSL_get_peer_certificate(r->connection->client->ssl);
+	if(xs)
+	    {
+	    s=X509_NAME_oneline(X509_get_subject_name(xs),NULL,0);
+	    r->connection->client->szClientX509=ap_pstrdup(r->connection->pool,
+							   s);
+	    free(s);
+	    }
+	}
+
+    if(rec->nVerifyClient == VERIFY_REQUIRE
+       && !r->connection->client->szClientX509)
+	{
+	SSLLogError(r->connection->server,APLOG_ERR,"No client certificate");
+	return FORBIDDEN;
+	}
+#endif
+
     cipher=SSL_get_cipher(r->connection->client->ssl);
 
     if(ap_table_get(rec->tbBannedCiphers,cipher))
@@ -305,22 +527,6 @@ static void ExpandCert(pool *p,table *pEnv,char *szPrefix,char *szCert)
 	}
     }
 
-#if APACHE_SSL_EXPORT_CERTS
-/* szEnv must have been pool allocated, or be static */
-static void AddCertToEnv(pool *p,table *e,char *szEnv,X509 *xs)
-    {
-    int n;
-    unsigned char *s,*d;
-
-    n=i2d_X509(xs,NULL);
-    d=s=ap_palloc(p,n);
-    i2d_X509(xs,&d);
-    d=ap_palloc(p,(n*4)/3+2+2);
-    uuencoden(d,s,n,TRUE);
-    ap_table_setn(e,szEnv,(char *)d);
-    }
-#endif
-
 static struct
     {
     char *szName;
@@ -421,7 +627,12 @@ static int SSLFixups(request_rec *r)
 	ExpandCert(r->pool,e,"SSL_CLIENT_",
 		   r->connection->client->szClientX509);
 	xs=SSL_get_peer_certificate(r->connection->client->ssl);
-	ExpandCert(r->pool,e,"SSL_CLIENT_I_",
+	ExpandCert(r->pool,e,
+#if USE_OLD_ENVIRONMENT
+		   "SSL_CLIENT_I",
+#else
+		   "SSL_CLIENT_I_",
+#endif
 		   X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
 #if APACHE_SSL_EXPORT_CERTS
 	if(pDirConfig->bExportCerts)
@@ -432,7 +643,12 @@ static int SSLFixups(request_rec *r)
     xs=SSL_get_certificate(r->connection->client->ssl);
     ExpandCert(r->pool,e,"SSL_SERVER_",
 	       X509_NAME_oneline(X509_get_subject_name(xs),NULL,0));
-    ExpandCert(r->pool,e,"SSL_SERVER_I_",
+    ExpandCert(r->pool,e,
+#if USE_OLD_ENVIRONMENT
+	       "SSL_SERVER_I",
+#else
+	       "SSL_SERVER_I_",
+#endif
 	       X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
 
 #if APACHE_SSL_EXPORT_CERTS
@@ -455,36 +671,6 @@ static int SSLFixups(request_rec *r)
     return DECLINED;
     }
 
-static void SSLLogError(server_rec *s,int nLevel,const char *szErr)
-    {
-    unsigned long l;
-    char buf[MAX_STRING_LEN];
-
-    if(szErr)
-	ap_log_error(APLOG_MARK,nLevel|APLOG_NOERRNO,s,"%s",szErr);
-    while((l=ERR_get_error()))
-	{
-	ERR_error_string(l,buf);
-	ap_log_error(APLOG_MARK,nLevel|APLOG_NOERRNO,s,"%s",buf);
-	}
-    }
-
-/* for debugging */
-static void SSLLogErrorToFile(FILE *f)
-    {
-    unsigned long l;
-    char buf[MAX_STRING_LEN];
-
-    /* Print out error messages */
-
-    while((l=ERR_get_error()))
-	{
-	ERR_error_string(l,buf);
-	fputs(buf,f);
-	fputc('\n',f);
-	}
-    }
-
 static void *CreateSSLServerConfig(pool *p,server_rec *s)
     {
     SSLConfigRec *rec=ap_pcalloc(p,sizeof(SSLConfigRec));
@@ -495,8 +681,10 @@ static void *CreateSSLServerConfig(pool *p,server_rec *s)
     rec->szCACertificatePath=NULL;
     rec->szCACertificateFile=NULL;
     rec->szReqCiphers=NULL;
+#if !RENEG
     rec->nVerifyDepth=0;
     rec->nVerifyClient=VERIFY_UNSET;	/* Will become VERIFY_NONE */
+#endif
     rec->px509Certificate=NULL;
     rec->prsaKey=NULL;
     rec->pSSLCtx=NULL;
@@ -522,8 +710,10 @@ static void *MergeSSLServerConfig(pool *p,void *basev,void *addv)
     MergeString(szKeyFile);
     MergeString(szCACertificatePath);
     MergeString(szCACertificateFile);
+#if !RENEG
     MergeInt(nVerifyDepth);
     Merge(nVerifyClient,VERIFY_UNSET);
+#endif
     MergeBool(bFakeBasicAuth);
     MergeInt(tSessionCacheTimeout);
 
@@ -540,6 +730,10 @@ static void *CreateSSLDirConfig(pool *p,char *dummy)
 #if APACHE_SSL_EXPORT_CERTS
     rec->bExportCerts=FALSE;
 #endif
+#if RENEG
+    rec->nVerifyDepth=0;
+    rec->nVerifyClient=VERIFY_UNSET;	/* Won't become VERIFY_NONE */
+#endif
 
     return rec;
     }
@@ -559,6 +753,11 @@ static void *MergeSSLDirConfig(pool *p,void *basev,void *addv)
 #if APACHE_SSL_EXPORT_CERTS
     new->bExportCerts=base->bExportCerts || add->bExportCerts;
 #endif
+#if RENEG
+    MergeInt(nVerifyDepth);
+    Merge(nVerifyClient,VERIFY_UNSET);
+#endif
+
     return new;
     }
 
@@ -606,122 +805,6 @@ static RSA *TmpRSACallback(SSL *pSSL,int nExport)
     }
 #endif
 
-#if SSLEAY_VERSION_NUMBER >= 0x0800
-static int ApacheSSLVerifyCallback(int ok,X509_STORE_CTX *ctx)
-    {
-    X509 *xs=X509_STORE_CTX_get_current_cert(ctx);
-    int depth=X509_STORE_CTX_get_error_depth(ctx);
-    int error=X509_STORE_CTX_get_error(ctx);
-#else
-static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
-				   int error)
-    {
-#endif
-    SSLConfigRec *pConfig=ap_get_module_config(s_pCurrentConnection->server->module_config,&apache_ssl_module);
-    server_rec *pServer=s_pCurrentConnection->server;
-    pool *p=s_pCurrentConnection->client->pool;
-
-#if DEBUG_VERIFY
-      {
-      char *s;
-
-#if SSLEAY_VERSION_NUMBER < 0x0800
-      s=X509_NAME_oneline(X509_get_subject_name(xs));
-#else
-      s=X509_NAME_oneline(X509_get_subject_name(xs),NULL,0);
-#endif
-
-      if(s == NULL)
-	  {
-	  SSLLogError(pServer,APLOG_ERR,"can't get subject name");
-	  return(0);
-	  }
-
-      ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"depth=%d %s",
-		   depth,s);
-      free(s);
-      }
-#endif
-    
-#if APACHE_SSL_KEEP_CERTS
-    if(depth)
-	{
-	char *t;
-
-	if(!s_pCurrentConnection->client->tableCertChain)
-	    s_pCurrentConnection->client->tableCertChain=ap_make_table(p,5);
-
-	t=ap_psprintf(p,ENV_CERT_CHAIN "%d",depth);
-	AddCertToEnv(p,s_pCurrentConnection->client->tableCertChain,t,xs);
-	}
-#endif
-
-    if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
-       || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE
-       || error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
-       || error == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN)
-	{
-	char *s;
-
-	if(pConfig->nVerifyClient == VERIFY_OPTIONAL_NO_CA)
-	    {
-	    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
-			 "no issuer, returning OK");
-	    return TRUE;
-	    }
-#if SSLEAY_VERSION_NUMBER < 0x0800
-	s=X509_NAME_oneline(X509_get_issuer_name(xs));
-#else
-	s=X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0);
-#endif
-	if(s == NULL)
-	    {
-	    SSLLogError(pServer,APLOG_ERR,"verify error");
-	    return FALSE;
-	    }
-	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"issuer=%s",
-		     s);
-	free(s);
-	}
-    if(!ok)
-	{
-	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,pServer,
-		      "verify error:num=%d:%s",error,
-		      X509_verify_cert_error_string(error));
-	SSLLogError(s_pCurrentConnection->server,APLOG_ERR,NULL);
-	s_pCurrentConnection->client->szClientX509=NULL;
-	}
-    if(depth >= pConfig->nVerifyDepth)
-	{
-	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,
-		     pServer,"Verify depth exceeded");
-	ok=0;
-	}
-    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
-		 "verify return:%d",ok);
-    return ok;
-    }
-
-static int VerifyFlags(SSLConfigRec *pConfig)
-    {
-    int nVerify=0;
-
-    switch(pConfig->nVerifyClient)
-	{
-    case VERIFY_REQUIRE:
-	nVerify|=SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
-	break;
-
-    case VERIFY_OPTIONAL:
-    case VERIFY_OPTIONAL_NO_CA:
-	nVerify|=SSL_VERIFY_PEER;
-	break;
-
-    default: break; /* avoid warnings from gcc */
-	}
-    return nVerify;
-    }
-
 #if CACHE_SESSIONS
 
 #define MAX_SESSION_DER	100000
@@ -764,9 +847,7 @@ static uchar *CacheFind(uchar *aucKey,int nKey,int *pnData,BOOL *pbLocal)
     return t;
     }
 
-static uchar *CacheFindS(char *aucKey,int nKey,int *pnData,BOOL *pbLocal)
-    { return CacheFind((uchar *)aucKey,nKey,pnData,pbLocal); }
-
+#if APACHE_SSL_EXPORT_CERTS
 static uchar *MakeCertKey(Cache *pCache,const char *szKey,int *pnLength)
     {
     uchar *buf=malloc(*pnLength=pCache->nKey+strlen(szKey));
@@ -792,7 +873,9 @@ static int SendCertToServer(void *extra,const char *szKey,const char *szValue)
     return 1;
     }
 
-#if APACHE_SSL_EXPORT_CERTS
+static uchar *CacheFindS(char *aucKey,int nKey,int *pnData,BOOL *pbLocal)
+    { return CacheFind((uchar *)aucKey,nKey,pnData,pbLocal); }
+
 static void SendCertChainToServer(Cache *pCache)
     {
     if(!s_pCurrentConnection->client->tableCertChain)
@@ -978,7 +1061,12 @@ static void GetPrivateKey(server_rec *s,SSLConfigRec *pConfig,
 		  (const char *)pConfig->prsaKey);
     }
 
-static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
+static int X509NameCompare(X509_NAME **a,X509_NAME **b)
+    {
+    return(X509_NAME_cmp(*a,*b));
+    }
+
+static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
     {
     FILE *f;
     char szPath[MAX_STRING_LEN];
@@ -995,7 +1083,12 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
     pConfig->pSSLCtx=SSL_CTX_new();
 #else
     pConfig->pSSLCtx=SSL_CTX_new(SSLv23_server_method());
+# if RENEG
+//    s_nVerifyClient=VERIFY_OPTIONAL_NO_CA;
+    SSL_CTX_set_verify(pConfig->pSSLCtx,0,ApacheSSLVerifyCallback);
+# else
     SSL_CTX_set_verify(pConfig->pSSLCtx,VerifyFlags(pConfig),ApacheSSLVerifyCallback);
+# endif
 #endif
 
 #if CACHE_SESSIONS
@@ -1070,12 +1163,48 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 #if SSLEAY_VERSION_NUMBER >= 0x0800
     SSL_CTX_set_tmp_rsa_callback(pConfig->pSSLCtx,TmpRSACallback);
 
-    /* Really this should have its own directive and list, but I haven't got
-       the time at the moment.
-     */
     if(pConfig->szCACertificateFile)
 	SSL_CTX_set_client_CA_list(pConfig->pSSLCtx,
-				   SSL_load_client_CA_file(pConfig->szCACertificateFile)); 
+				   SSL_load_client_CA_file(pConfig->szCACertificateFile));
+
+    if(pConfig->szCACertificatePath)
+	{
+	pool *p2=ap_make_sub_pool(p);
+	DIR *d=ap_popendir(p2,pConfig->szCACertificatePath);
+	struct DIR_TYPE *dstruct;
+	STACK *stackCAs;
+
+	/* Note that a side effect is that the CAs will be sorted by name */
+	stackCAs=sk_new(X509NameCompare);
+	if(!d)
+	    {
+	    ap_log_error(APLOG_MARK,APLOG_ERR,s,"Can't open directory %s",
+			 pConfig->szCACertificatePath);
+	    exit(1);
+	    }
+
+	while((dstruct=readdir(d)))
+	    {
+	    STACK *s;
+	    int n;
+	    char *szFile;
+	    
+	    szFile=ap_pstrcat(p2,pConfig->szCACertificatePath,"/",
+			      dstruct->d_name,NULL);
+	    s=SSL_load_client_CA_file(szFile);
+	    for(n=0 ; s && n < sk_num(s) ; ++n)
+		{
+		if(sk_find(stackCAs,sk_value(s,n)) >= 0)
+		    continue;
+	        sk_push(stackCAs,sk_value(s,n));
+		}
+	    }
+
+	sk_set_cmp_func(stackCAs,NULL);
+	SSL_CTX_set_client_CA_list(pConfig->pSSLCtx,stackCAs);
+	ap_pclosedir(p2,d);
+	ap_destroy_pool(p2);
+	}
 #endif
     }
 
@@ -1151,8 +1280,10 @@ static void InitSSLServer(server_rec *s,pool *p)
 
 	if(pConfig->bDisabled == UNSET)
 	    pConfig->bDisabled=FALSE;
+#if !RENEG
 	if(pConfig->nVerifyClient == VERIFY_UNSET)
 	    pConfig->nVerifyClient=VERIFY_NONE;
+#endif
 	if(pConfig->bFakeBasicAuth == UNSET)
 	    pConfig->bFakeBasicAuth=FALSE;
 
@@ -1185,6 +1316,7 @@ static void InitSSLServer(server_rec *s,pool *p)
 			 s->server_hostname,s->port);
 	    exit(1);
 	    }
+#if !RENEG
 	if(pConfig->nVerifyClient < 0 || pConfig->nVerifyClient > VERIFY_OPTIONAL_NO_CA)
 	    {
 	    ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
@@ -1192,7 +1324,8 @@ static void InitSSLServer(server_rec *s,pool *p)
 			 pConfig->nVerifyClient);
 	    exit(1);
 	    }
-	GetCertificateAndKey(s,pConfig);
+#endif
+	GetCertificateAndKey(s,p,pConfig);
 	}
 #if CACHE_SESSIONS
     LaunchCache(pMainServer,p);
@@ -1200,6 +1333,7 @@ static void InitSSLServer(server_rec *s,pool *p)
     close(2);
     n=dup(nSaveStderr);
     assert(n == 2);
+    close(nSaveStderr);
     }
 
 /*
@@ -1229,15 +1363,15 @@ static int FakeBasicAuth(request_rec *r)
     if(!r->connection->client->szClientX509)
 	return DECLINED;
     /*
-       Fake a password - which one would be immaterial, as, it seems, an empty
-       password in the users file would match ALL incoming passwords, if only we
-       were using the standard crypt library routine. Unfortunately, SSLeay
-       "fixes" a "bug" in crypt and thus prevents blank passwords from working.
-       (IMHO what they really fix is a bug in the users of the code - failing to
-       program correctly for shadow passwords).
-       We need, therefore, to provide a password. This password can be matched by
-       adding the string "xxj31ZMTZzkVA" as the password in the user file.
-*/
+      Fake a password - which one would be immaterial, as, it seems, an empty
+      password in the users file would match ALL incoming passwords, if only we
+      were using the standard crypt library routine. Unfortunately, SSLeay
+      "fixes" a "bug" in crypt and thus prevents blank passwords from working.
+      (IMHO what they really fix is a bug in the users of the code - failing to
+      program correctly for shadow passwords).
+      We need, therefore, to provide a password. This password can be matched
+      by adding the string "xxj31ZMTZzkVA" as the password in the user file.
+      */
     ap_snprintf(b1,sizeof b1,"%s:password",
 		r->connection->client->szClientX509);
     uuencode(b2,b1,FALSE);
@@ -1308,6 +1442,9 @@ static void CloseSSL(void *_conn)
     SSL_free(ssl);
     /*    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,conn->server,
 		 "Closed");*/
+#if RENEG
+    s_pCurrentConnection=NULL;
+#endif
     }
 #endif
 
@@ -1317,12 +1454,13 @@ int ApacheSSLSetupConnection(conn_rec * conn)
     BUFF *fb=conn->client;
     SSLConfigRec *pConfig=ap_get_module_config(srvr->module_config,
 					       &apache_ssl_module);
-
     if(pConfig->bDisabled)
 	{
 	fb->ssl=NULL;
 	return TRUE;
 	}
+
+    ap_conn_timeout("SSL accept timeout",conn);
     
     s_pCurrentConnection=conn;
 
@@ -1340,7 +1478,17 @@ int ApacheSSLSetupConnection(conn_rec * conn)
     while(!SSL_is_init_finished(fb->ssl))
 	{
 	int ret=SSL_accept(fb->ssl);
-	if (ret <= 0)
+	if(ret == 0)
+	    {
+	    /* This appears to happen when a connection is closed without any
+	       data being sent. Since Netscape does this all the time, don't
+	       log as an error.
+	       */
+	    SSLLogError(srvr,APLOG_DEBUG,"SSL_accept returned 0");
+	    fb->flags |= B_EOF | B_EOUT;
+	    return FALSE;
+	    }
+	if (ret < 0)
 	    {
 	    SSLLogError(srvr,APLOG_ERR,"SSL_accept failed");
 	    fb->flags |= B_EOF | B_EOUT;
@@ -1354,6 +1502,7 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 	    return FALSE;
 	    }
 
+#if !RENEG
 	if(pConfig->nVerifyClient != VERIFY_NONE)
 	    {
 	    char *s;
@@ -1372,14 +1521,17 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 	    SSLLogError(conn->server,APLOG_ERR,"No client certificate");
 	    return 0;
 	    }
+#endif
 	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,srvr,"CIPHER is %s",
 		     SSL_get_cipher(conn->client->ssl));
 	}
 
+#if !RENEG
     /* This should be safe.... so I'll use it */
     SSL_set_read_ahead(fb->ssl,1);
 
     s_pCurrentConnection=NULL;
+#endif
 
     return TRUE;
     }
@@ -1389,6 +1541,17 @@ static const char *SSLLogFile(cmd_parms *cmd,char *struct_ptr,char *arg)
     return "The SSLLogFile directive is no longer used. All logging information now goes to the error log.";
     }
 
+#if RENEG
+static const char *SSLVerifyClient(cmd_parms *cmd,SSLDirConfigRec *rec,
+				   char *level)
+    {
+    rec->nVerifyClient=atoi(level);
+    if(rec->nVerifyClient < 0 || rec->nVerifyClient > 3)
+	return "Bad value for SSLVerifyClient";
+    return NULL;
+    }
+#endif
+
 static const char *SSLRequireCipher(cmd_parms *cmd,SSLDirConfigRec *rec,
 				    char *cipher)
     {
@@ -1449,9 +1612,9 @@ static const char *SSLCacheServerDir(cmd_parms *cmd,char *struct_ptr,
     }
 #endif
 
-static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,char *arg)
+static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,
+					   char *arg)
     {
-    /* This one should be pretty generic... */
     char *pConfig=ap_get_module_config(cmd->server->module_config,
 				       &apache_ssl_module);
   
@@ -1460,9 +1623,18 @@ static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,char
     return NULL;
     }
 
-static const char *set_server_int_slot (cmd_parms *cmd,char *struct_ptr,char *arg)
+#if RENEG
+static const char *set_dir_int_slot(cmd_parms *cmd,char *pStruct,char *szValue)
+    {
+    int offset=(int)(long)cmd->info;
+    *(int *)(pStruct+offset)=atoi(szValue);
+    return NULL;
+    }
+#endif
+
+static const char *set_server_int_slot (cmd_parms *cmd,char *struct_ptr,
+					char *arg)
     {
-  /* This one should be pretty generic... */
     char *pConfig=ap_get_module_config(cmd->server->module_config,
 				       &apache_ssl_module);
   
@@ -1473,7 +1645,6 @@ static const char *set_server_int_slot (cmd_parms *cmd,char *struct_ptr,char *ar
 
 static const char *set_server_bool_slot (cmd_parms *cmd,char *struct_ptr)
     {
-  /* This one should be pretty generic... */
     char *pConfig=ap_get_module_config(cmd->server->module_config,
 				       &apache_ssl_module);
   
@@ -1484,7 +1655,6 @@ static const char *set_server_bool_slot (cmd_parms *cmd,char *struct_ptr)
 
 static const char *set_server_bool_slot_false (cmd_parms *cmd,char *struct_ptr)
     {
-  /* This one should be pretty generic... */
     char *pConfig=ap_get_module_config(cmd->server->module_config,
 				       &apache_ssl_module);
   
@@ -1513,12 +1683,20 @@ command_rec ssl_cmds[]=
     { "SSLCACertificateFile",set_server_string_slot,
       (void *)XtOffsetOf(SSLConfigRec,szCACertificateFile),RSRC_CONF,TAKE1,
       "CA Certificate file (taken from SSL_CERT_FILE if absent)" },
+#if RENEG
+    { "SSLVerifyDepth",set_dir_int_slot,
+      (void *)XtOffsetOf(SSLDirConfigRec,nVerifyDepth),OR_FILEINFO,TAKE1,
+      "Verify depth (default 0)" },
+    { "SSLVerifyClient",SSLVerifyClient,NULL,OR_FILEINFO,TAKE1,
+      "Verify client (0=no,1=optional,2=required,3=required (CA optional)" },
+#else
     { "SSLVerifyDepth",set_server_int_slot,
       (void *)XtOffsetOf(SSLConfigRec,nVerifyDepth),RSRC_CONF,TAKE1,
       "Verify depth (default 0)" },
     { "SSLVerifyClient",set_server_int_slot,
       (void *)XtOffsetOf(SSLConfigRec,nVerifyClient),RSRC_CONF,TAKE1,
       "Verify client (0=no,1=optional,2=required" },
+#endif
     { "SSLFakeBasicAuth",set_server_bool_slot,
       (void *)XtOffsetOf(SSLConfigRec,bFakeBasicAuth),RSRC_CONF,NO_ARGS,
       "Translate client X509 into a user name" },
